/*
 * Loesung zu Blatt 11
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <string.h>

char *opt_i = NULL;	/* Option -i gesetzt? => Argument */
int inode;
char *opt_n = NULL;	/* Option -n gesetzt? => Argument */
char *opt_a = NULL;	/* Option -a gesetzt? => Argument */
int atime;
int opt_r = 0;		/* Option -r gesetzt? */
int opt_w = 0;		/* Option -w gesetzt? */
int opt_x = 0;		/* Option -x gesetzt? */
int opt_l = 0;		/* Option -l gesetzt? */

/*
 * Wandelt String s in eine Integer um,
 * die in *val geschrieben wird.
 * Liefert im Erfolgsfall true (!= 0) und
 * sonst false (0).
 */
int to_int(char *s, int *val) {
    char *p;
    errno = 0;	/* errno zuerst auf 0 setzen */
    /*
     * liefert die Zahl, die der String s enthaelt;
     * p zeigt danach auf die Stelle im String nach
     * dieser Zahl
     */
    *val = strtol(s, &p, 10);
    /*
     * Fehler, wenn errno != 0 ist oder aber
     * p noch nicht auf das Stringende (also Null-Byte) zeigt.
     */
    if (errno != 0 || *p != '\0')
	return 0;	/* FEHLER */
    return 1;		/* alles ok */
}

/*
 * In allen Unterverzeichnissen von path rekursiv suchen.
 */
void recurse_subdirs(char *path) {
    DIR *dir;
    struct dirent *entry;

    /* Verbindung zum Verzeichnis oeffnen */
    if (!(dir = opendir(path))) {
	char err[1024];
	sprintf(err, "opendir: %s", path);
	perror(err);
	return;
    }

    /* alle Verzeichniseintraege durchlaufen ... */
    while ((entry = readdir(dir))) {
	char buf[1024];
	struct stat statbuf;

	/* bei . und .. machen wir nichts und gehen gleich zum
	 * naechsten Verzeichniseintrag */
	if (!strcmp(entry->d_name, ".") || !strcmp(entry->d_name, ".."))
	    continue;

	/* Pfadnamen zusammenbauen aus Pfad + Dateiname */
	sprintf(buf, "%s/%s", path, entry->d_name);

	/* je nachdem, ob -l gesetzt bzw. nicht gesetzt,
	 * verwenden wir stat() bzw. lstat() */
	if (opt_l) {
	    if (stat(buf, &statbuf) < 0) {
		char err[1024];
		sprintf(err, "stat: %s", buf);
		perror(err);
		continue;
	    }
	}
	else {
	    if (lstat(buf, &statbuf) < 0) {
		char err[1024];
		sprintf(err, "lstat: %s", buf);
		perror(err);
		continue;
	    }
	}

	/* fuer die Ausgabe einer Datei muss folgendes fuer jede
	 * Option gelten: Entweder die Option ist nicht angegeben ;-),
	 * oder die Bedingung, die zu dieser Option gehoert, ist
	 * fuer diese Datei erfuellt ... und dies verknuepfen wird
	 * mit UND fuer alle Optionen */
	if ((!opt_i || statbuf.st_ino == inode)			/* Inode-Nummer = inode */
	    && (!opt_n || !strcmp(entry->d_name, opt_n))	/* Dateiname = opt_n (Optionsarg.) */
	    && (!opt_a || atime < statbuf.st_atime)		/* Zugriffszeit neuer als atime */
	    && (!opt_r || (statbuf.st_mode & 0400) != 0)	/* Besitzer der Datei kann lesen */
	    && (!opt_w || (statbuf.st_mode & 0200) != 0)	/* Besitzer der Datei kann schreiben */
	    && (!opt_x || (statbuf.st_mode & 0100) != 0))	/* Besitzer der Datei kann ausfuehren */
	    puts(buf);

	/* ... wenn's sich um ein Verzeichnis handelt
	 * (aber ja nicht mehr . und ..), dann steigen
	 * wir rekursiv hinunter ... */
	if (S_ISDIR(statbuf.st_mode))
	    recurse_subdirs(buf);
    }

    /* ... und die Verbindung zum Verzeichnis sauber beenden */
    closedir(dir);
}

int main(int argc, char *argv[]) {
    char *cmdname = argv[0];	/* Kommando-Name */
    int opterr = 0;		/* Fehler bei den Optionen? */
    int atend;			/* Scannen eines Options-Parameters abbrechen? */
    char *s;

	/* Optionen verarbeiten */
    while (--argc > 0 && (*++argv)[0] == '-') {
	/* nach "--" kommen keine Optionen mehr (per Definition) */
	if ((*argv)[1] == '-' && (*argv)[2] == '\0') {
	    argc--; argv++; break;
	}
	/* steht nur "-", dann fehlt etwas ;-) */
	if ((*argv)[1] == '\0') {
	    fprintf(stderr, "Empty option\n");
	    opterr = 1;
	}
	for (s = *argv + 1, atend = 0; *s != '\0' && !atend; s++)
	    switch (*s) {	/* jede einzelne Option abarbeiten ... */
		case 'r' : opt_r = 1; break;	/* Option -r ohne Parameter behandeln */
		case 'w' : opt_w = 1; break;
		case 'x' : opt_x = 1; break;
		case 'l' : opt_l = 1; break;
		case 'i' :	/* Option -i mit Parameter behandeln */
		    if (opt_i) {
			fprintf(stderr, "Option '%c' occured several times\n", 'i');
			opterr = 1;
			atend = 1;
			break;
		    }
		    if (*(s+1) != '\0')		/* steht der Parameter zur Option direkt dahinter (ohne Leerzeichen) */
			opt_i = s+1;		/* opt_i ist Zeiger auf den Parameterwert */
		    else if (argc-1 > 0) {	/* ... ansonsten (nicht dahinter) gibts noch ein Kommandozeilenarg.? */
			argc--;			/* dieses Kommandozeilenarg. verbrauchen ... */
			opt_i = *++argv;	/* ... und als Parameter von -i in opt_i speichern */
		    }
		    else {			/* andernfalls gab es kein weiteres Kommandozeilenarg. mehr => Fehler */
			fprintf(stderr, "No argument for option '%c' present\n", 'i');
			opterr = 1;		/* bewirkt die Ausgabe der Usage-Meldung nach der Bearbeitung der Optionen */
		    }
		    atend = 1;			/* bewirkt das Ende der direkt umgebenden for-Schleife (=> dieses Kommandozeilenarg. ist bearbeitet) */
		    break;
		case 'n' :	/* ... analog zu -i */
		    if (*(s+1) != '\0')
			opt_n = s+1;
		    else if (argc-1 > 0) {
			argc--;
			opt_n = *++argv;
		    }
		    else {
			fprintf(stderr, "No argument for option '%c' present\n", 'n');
			opterr = 1;
		    }
		    atend = 1;
		    break;
		case 'a' :	/* ... analog zu -i */
		    if (*(s+1) != '\0')
			opt_a = s+1;
		    else if (argc-1 > 0) {
			argc--;
			opt_a = *++argv;
		    }
		    else {
			fprintf(stderr, "No argument for option '%c' present\n", 'a');
			opterr = 1;
		    }
		    atend = 1;
		    break;
		default:	/* unbekannte Option! */
		    fprintf(stderr, "Illegal option '%c'\n", *s);
		    opterr = 1;
		    atend = 1;
		    break;
	    }
    }

    /*
     * beachte: Durch die Optionsverarbeitung wurden argc und argv veraendert!
     */

    /* Fehler bei zuvielen Argumenten ausgeben */
    if (argc > 1)
	fprintf(stderr, "Too many arguments\n");

    /* Usage-Meldung, falls zuviele Argumente, oder Fehler bei den Optionen */
    if (argc > 1 || opterr) {
	fprintf(stderr, "Usage: %s [-i <inode>] [-n <file>] [-a <file>] [-r|-w|-x|-l] [<directory>]\n", cmdname);
	return 1;
    }

    /* falls Option -i, dann noch den String opt_i parsen => muss Integer >= 0 sein! */
    if (opt_i && (!to_int(opt_i, &inode) || inode < 0)) {
	fprintf(stderr, "Argument of -i is not a non-negative integer: %s\n", opt_i);
	return 2;
    }

    /* Option -a => Zugriffszeit der Datei ermitteln */
    if (opt_a) {
	struct stat statbuf;
	/* meine Wahl: lstat(); stat() waere auch okay */
	if (lstat(opt_a, &statbuf) < 0) {
	    char err[1024];
	    sprintf(err, "lstat: %s", opt_a);
	    perror(err);
	    return 3;
	}
	/* Zugriffszeit der Datei mit dem Namen, der bei -a als Argument angegeben wurde */
	atime = (int) statbuf.st_atime;
    }

    /* jetzt suchen wir rekrusiv in den Unterverzeichnissen
     * von ., falls kein Argument, und sonst im angegebenen Verzeichnis */
    recurse_subdirs(argc > 0 ? *argv : ".");

    return 0;
}
